@@ -65,7 +65,10 @@
 #' @param .params Additional list of parameters to append to `...`.
 #'   It is easier to use this than `...` if you have your parameters in
 #'   a list already.
-#'
+#' @param .max_wait Maximum number of seconds to wait if rate limited.
+#'   Defaults to 10 minutes.
+#' @param .max_rate Maximum request rate in requests per second. Set
+#'   this to automatically throttle requests.
 #' @return Answer from the API as a `gh_response` object, which is also a
 #'   `list`. Failed requests will generate an R error. Requests that
 #'   generate a raw response will return a raw vector.
@@ -75,18 +78,16 @@
 #' [gh_whoami()] for details on GitHub API token management.
 #' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
 #' ## Repositories of a user, these are equivalent
-#' gh("/users/hadley/repos")
-#' gh("/users/{username}/repos", username = "hadley")
+#' gh("/users/hadley/repos", .limit = 2)
+#' gh("/users/{username}/repos", username = "hadley", .limit = 2)
 #'
 #' ## Starred repositories of a user
-#' gh("/users/hadley/starred")
-#' gh("/users/{username}/starred", username = "hadley")
-#'
+#' gh("/users/hadley/starred", .limit = 2)
+#' gh("/users/{username}/starred", username = "hadley", .limit = 2)
 #' @examplesIf FALSE
 #' ## Create a repository, needs a token in GITHUB_PAT (or GITHUB_TOKEN)
 #' ## environment variable
 #' gh("POST /user/repos", name = "foobar")
-#'
 #' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
 #' ## Issues of a repository
 #' gh("/repos/hadley/dplyr/issues")
@@ -95,24 +96,20 @@
 #' ## Automatic pagination
 #' users <- gh("/users", .limit = 50)
 #' length(users)
-#'
 #' @examplesIf FALSE
 #' ## Access developer preview of Licenses API (in preview as of 2015-09-24)
 #' gh("/licenses") # used to error code 415
 #' gh("/licenses", .accept = "application/vnd.github.drax-preview+json")
-#'
 #' @examplesIf FALSE
 #' ## Access Github Enterprise API
 #' ## Use GITHUB_API_URL environment variable to change the default.
 #' gh("/user/repos", type = "public", .api_url = "https://github.foobar.edu/api/v3")
-#'
 #' @examplesIf FALSE
 #' ## Use I() to force body part to be sent as an array, even if length 1
 #' ## This works whether assignees has length 1 or > 1
 #' assignees <- "gh_user"
 #' assignees <- c("gh_user1", "gh_user2")
 #' gh("PATCH /repos/OWNER/REPO/issues/1", assignees = I(assignees))
-#'
 #' @examplesIf FALSE
 #' ## There are two ways to send JSON data. One is that you supply one or
 #' ## more objects that will be converted to JSON automatically via
@@ -124,11 +121,11 @@
 #'
 #' gh::gh(
 #'   "POST /repos/{owner}/{repo}/pages",
-#'   owner = "gaborcsardi",
-#'   repo = "playground",
+#'   owner = "r-lib",
+#'   repo = "gh",
 #'   source = list(
-#'     branch = jsonlite::unbox("master"),
-#'     path = jsonlite::unbox("/docs")
+#'     branch = jsonlite::unbox("gh-pages"),
+#'     path = jsonlite::unbox("/")
 #'   ),
 #'   .send_headers = c(Accept = "application/vnd.github.switcheroo-preview+json")
 #' )
@@ -136,22 +133,32 @@
 #' ## The second way is to handle the JSON encoding manually, and supply it
 #' ## as a raw vector in an unnamed argument, and also a Content-Type header:
 #'
-#' body <- '{ "source": { "branch": "master", "path": "/docs" } }'
+#' body <- '{ "source": { "branch": "gh-pages", "path": "/" } }'
 #' gh::gh(
 #'   "POST /repos/{owner}/{repo}/pages",
-#'   owner = "gaborcsardi",
-#'   repo = "playground",
+#'   owner = "r-lib",
+#'   repo = "gh",
 #'   charToRaw(body),
 #'   .send_headers = c(
 #'     Accept = "application/vnd.github.switcheroo-preview+json",
 #'     "Content-Type" = "application/json"
 #'   )
 #' )
-gh <- function(endpoint, ..., per_page = NULL, .token = NULL, .destfile = NULL,
-               .overwrite = FALSE, .api_url = NULL, .method = "GET",
-               .limit = NULL, .accept = "application/vnd.github.v3+json",
-               .send_headers = NULL, .progress = TRUE, .params = list()) {
-
+gh <- function(endpoint,
+               ...,
+               per_page = NULL,
+               .token = NULL,
+               .destfile = NULL,
+               .overwrite = FALSE,
+               .api_url = NULL,
+               .method = "GET",
+               .limit = NULL,
+               .accept = "application/vnd.github.v3+json",
+               .send_headers = NULL,
+               .progress = TRUE,
+               .params = list(),
+               .max_wait = 600,
+               .max_rate = NULL) {
   params <- c(list(...), .params)
   params <- drop_named_nulls(params)
 
@@ -165,20 +172,26 @@ gh <- function(endpoint, ..., per_page = NULL, .token = NULL, .destfile = NULL,
     params <- c(params, list(per_page = per_page))
   }
 
-  req <- gh_build_request(endpoint = endpoint, params = params,
-                          token = .token, destfile = .destfile,
-                          overwrite = .overwrite, accept = .accept,
-                          send_headers = .send_headers,
-                          api_url = .api_url, method = .method)
-
+  req <- gh_build_request(
+    endpoint = endpoint,
+    params = params,
+    token = .token,
+    destfile = .destfile,
+    overwrite = .overwrite,
+    accept = .accept,
+    send_headers = .send_headers,
+    max_wait = .max_wait,
+    max_rate = .max_rate,
+    api_url = .api_url,
+    method = .method
+  )
 
   if (req$method == "GET") check_named_nas(params)
 
   if (.progress) prbr <- make_progress_bar(req)
 
   raw <- gh_make_request(req)
-
-  res <- gh_process_response(raw)
+  res <- gh_process_response(raw, req)
   len <- gh_response_length(res)
 
   while (!is.null(.limit) && len < .limit && gh_has_next(res)) {
@@ -186,8 +199,8 @@ gh <- function(endpoint, ..., per_page = NULL, .token = NULL, .destfile = NULL,
     res2 <- gh_next(res)
 
     if (!is.null(names(res2)) && identical(names(res), names(res2))) {
-      res3 <- mapply(           # Handle named array case
-        function(x, y, n) {        # e.g. GET /search/repositories
+      res3 <- mapply( # Handle named array case
+        function(x, y, n) { # e.g. GET /search/repositories
           z <- c(x, y)
           atm <- is.atomic(z)
           if (atm && n %in% c("total_count", "incomplete_results")) {
@@ -201,8 +214,8 @@ gh <- function(endpoint, ..., per_page = NULL, .token = NULL, .destfile = NULL,
         res, res2, names(res),
         SIMPLIFY = FALSE
       )
-    } else {                    # Handle unnamed array case
-      res3 <- c(res, res2)      # e.g. GET /orgs/:org/invitations
+    } else { # Handle unnamed array case
+      res3 <- c(res, res2) # e.g. GET /orgs/:org/invitations
     }
 
     len <- len + gh_response_length(res2)
@@ -212,8 +225,8 @@ gh <- function(endpoint, ..., per_page = NULL, .token = NULL, .destfile = NULL,
   }
 
   # We only subset for a non-named response.
-  if (! is.null(.limit) && len > .limit &&
-      ! "total_count" %in% names(res) && length(res) == len) {
+  if (!is.null(.limit) && len > .limit &&
+    !"total_count" %in% names(res) && length(res) == len) {
     res_attr <- attributes(res)
     res <- res[seq_len(.limit)]
     attributes(res) <- res_attr
@@ -224,7 +237,7 @@ gh <- function(endpoint, ..., per_page = NULL, .token = NULL, .destfile = NULL,
 
 gh_response_length <- function(res) {
   if (!is.null(names(res)) && length(res) > 1 &&
-      names(res)[1] == "total_count") {
+    names(res)[1] == "total_count") {
     # Ignore total_count, incomplete_results, repository_selection
     # and take the first list element to get the length
     lst <- vapply(res, is.list, logical(1))
@@ -233,20 +246,109 @@ gh_response_length <- function(res) {
       c("total_count", "incomplete_results", "repository_selection")
     )
     tgt <- which(lst[nm])[1]
-    if (is.na(tgt)) length(res) else length(res[[ nm[tgt] ]])
+    if (is.na(tgt)) length(res) else length(res[[nm[tgt]]])
   } else {
     length(res)
   }
 }
 
-gh_make_request <- function(x) {
+gh_make_request <- function(x, error_call = caller_env()) {
+  if (!x$method %in% c("GET", "POST", "PATCH", "PUT", "DELETE")) {
+    cli::cli_abort("Unknown HTTP verb: {.val {x$method}}")
+  }
 
-  method_fun <- list("GET" = GET, "POST" = POST, "PATCH" = PATCH,
-                     "PUT" = PUT, "DELETE" = DELETE)[[x$method]]
-  if (is.null(method_fun)) throw(new_error("Unknown HTTP verb"))
+  req <- httr2::request(x$url)
+  req <- httr2::req_method(req, x$method)
+  req <- httr2::req_url_query(req, !!!x$query)
+  if (is.raw(x$body)) {
+    req <- httr2::req_body_raw(req, x$body)
+  } else {
+    req <- httr2::req_body_json(req, x$body, null = "list", digits = 4)
+  }
+  req <- httr2::req_headers(req, !!!x$headers)
 
-  raw <- do.call(method_fun,
-                 compact(list(url = x$url, query = x$query, body = x$body,
-                              add_headers(x$headers), x$dest)))
-  raw
+  if (!is_testing()) {
+    req <- httr2::req_retry(
+      req,
+      max_tries = 3,
+      is_transient = function(resp) github_is_transient(resp, x$max_wait),
+      after = github_after
+    )
+  }
+
+  if (!is.null(x$max_rate)) {
+    req <- httr2::req_throttle(req, x$max_rate)
+  }
+
+  # allow custom handling with gh_error
+  req <- httr2::req_error(req, is_error = function(resp) FALSE)
+
+  resp <- httr2::req_perform(req, path = x$dest)
+  if (httr2::resp_status(resp) >= 300) {
+    gh_error(resp, error_call = error_call)
+  }
+
+  resp
+}
+
+# https://docs.github.com/v3/#client-errors
+gh_error <- function(response, error_call = caller_env()) {
+  heads <- httr2::resp_headers(response)
+  res <- httr2::resp_body_json(response)
+  status <- httr2::resp_status(response)
+
+  msg <- "GitHub API error ({status}): {heads$status %||% ''} {res$message}"
+
+  if (status == 404) {
+    msg <- c(msg, x = c("URL not found: {.url {response$url}}"))
+  }
+
+  doc_url <- res$documentation_url
+  if (!is.null(doc_url)) {
+    msg <- c(msg, c("i" = "Read more at {.url {doc_url}}"))
+  }
+
+  errors <- res$errors
+  if (!is.null(errors)) {
+    errors <- as.data.frame(do.call(rbind, errors))
+    nms <- c("resource", "field", "code", "message")
+    nms <- nms[nms %in% names(errors)]
+    msg <- c(
+      msg,
+      capture.output(print(errors[nms], row.names = FALSE))
+    )
+  }
+
+  cli::cli_abort(
+    msg,
+    class = c("github_error", paste0("http_error_", status)),
+    call = error_call,
+    response_headers = heads,
+    response_content = res
+  )
+}
+
+
+# use retry-after info when possible
+# https://docs.github.com/en/rest/overview/resources-in-the-rest-api#exceeding-the-rate-limit
+github_is_transient <- function(resp, max_wait) {
+  if (httr2::resp_status(resp) != 403) {
+    return(FALSE)
+  }
+  if (!identical(httr2::resp_header(resp, "x-ratelimit-remaining"), "0")) {
+    return(FALSE)
+  }
+
+  time <- httr2::resp_header(resp, "x-ratelimit-reset")
+  if (is.null(time)) {
+    return(FALSE)
+  }
+
+  time <- as.numeric(time)
+  minutes_to_wait <- (time - unclass(Sys.time()))
+  minutes_to_wait <= max_wait
+}
+github_after <- function(resp) {
+  time <- as.numeric(httr2::resp_header(resp, "x-ratelimit-reset"))
+  time - unclass(Sys.time())
 }
